from abc import ABC
from typing import List

class Selector(ABC):
    def __init__(self, to_embed:List[str] = [], to_metadata:List[str] = []):
        self.to_embed = to_embed
        self.to_metadata = to_metadata
    
    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['to_embed'] = self.to_embed
        json_to_return['to_metadata'] = self.to_metadata
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different han toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['to_embed'] = self.to_embed
        json_to_return['to_metadata'] = self.to_metadata
        return json_to_return

    def as_selector(dct:dict):
        if dct == None:
            return Selector(to_embed=[], to_metadata=[])
        return Selector(
            to_embed=dct.get("to_embed", []),
            to_metadata=dct.get("to_metadata", []),
        )


from abc import ABC
from starlette.exceptions import HTTPException

class NeumDocument(ABC):
    def __init__(self, id:str, content:str, metadata:dict) -> None:
        self.id:str = id
        self.content:str = content
        self.metadata:dict = metadata

    def as_file(dct:dict):
        if dct == None:
            raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
        return NeumDocument(
            content=dct.get("content", None),
            metadata=dct.get("metadata", None),
            id=dct.get("id", None),
        )
    
    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['content'] = self.content
        json_to_return['metadata'] = self.metadata
        json_to_return['id'] = self.id
        return json_to_return
    


from starlette.exceptions import HTTPException
from abc import ABC, abstractmethod
from typing import List, Tuple
from neumai.Shared.NeumDocument import NeumDocument

class EmbedConnector(ABC):
    def __init__(self, embed_information: dict = {}):
        self.embed_information = embed_information

    @property
    @abstractmethod
    def embed_name(self) -> str:
        pass
    
    @property
    @abstractmethod
    def requiredProperties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optionalProperties(self) -> List[str]:
        pass

    @abstractmethod
    def validation(self) -> bool:
        """Validate connector setup"""

    @abstractmethod
    def embed(self, documents:List[NeumDocument]) -> Tuple[List, dict]:
        """Generate embeddings with a given service"""
    
    @abstractmethod
    def embed_query(self, query:str) -> List[float]:
        """Generate embeddings with a given service"""

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['embed_name'] = self.embed_name
        json_to_return['embed_information'] = self.embed_information
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different han toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['embed_name'] = self.embed_name
        json_to_return['embed_information'] = self.embed_information
        return json_to_return
    
    def config(self):
        json_to_return = {}
        json_to_return['requiredProperties'] = self.requiredProperties
        json_to_return['optionalProperties'] = self.optionalProperties
        return json_to_return


from abc import abstractmethod, ABC, abstractproperty
from neumai.Chunkers.Chunker import Chunker
from starlette.exceptions import HTTPException
from neumai.Shared.NeumDocument import NeumDocument
from typing import List, Generator
from langchain.text_splitter import (CharacterTextSplitter)

class CharacterChunker(Chunker):
    """" Character Chunker \n
    chunker_information optional: \n
    [ chunk_size, chunk_overlap, batch_size ]"""
    
    @property
    def chunker_name(self) -> str:
        return "CharacterChunker"
    
    @property
    def requiredProperties(self) -> List[str]:
        return []

    @property
    def optionalProperties(self) -> List[str]:
        return ["chunk_size" , "chunk_overlap" , "batch_size", "separator"]

    def chunk(self, documents:List[NeumDocument]) -> Generator[List[NeumDocument], None, None]:

        batch_size = self.chunker_information.get('batch_size', 1000)

        text_splitter = CharacterTextSplitter(
            separator = self.chunker_information.get('separator', "\n\n"),
            chunk_size = self.chunker_information.get('chunk_size', 500),
            chunk_overlap = self.chunker_information.get('chunk_overlap', 0),
            length_function = len,
        )
        
        # Iterate through documents to chunk them and them merge them back up
        documents_to_embed:List[NeumDocument] = []
        for doc in documents:
            chunks = text_splitter.split_text(doc.content)
            for i in range(len(chunks)):
                documents_to_embed.append(NeumDocument(id=doc.id + "_" + str(i), content=chunks[i], metadata=doc.metadata))
                if(len(documents_to_embed) == batch_size):
                    yield documents_to_embed
                    documents_to_embed = []

        # Pass last remaining
        if(len(documents_to_embed) > 0):
            yield documents_to_embed

    def validate(self) -> bool:
        return True   


from datetime import datetime
from typing import List, Generator
import boto3
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from neumai.DataConnectors.DataConnector import DataConnector
import tempfile
import os


class S3Connector(DataConnector):
    """" Neum File Connector \n
    connector_information contains: \n
    [ aws_key_id, aws_access_key, bucket_name ]"""

    @property
    def connector_name(self) -> str:
        return "S3Connector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["aws_key_id", "aws_access_key", "bucket_name"]

    @property
    def optionalProperties(self) -> List[str]:
        return ["prefix"]
    
    @property
    def availableMetadata(self) -> str:
        return ["key" , "last_modified", "metadata" ] # TODO need to flatten

    @property
    def availableContent(self) -> str:
        return ['file']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        #Connect to S3
        aws_key_id= self.connector_information['aws_key_id']
        aws_access_key= self.connector_information['aws_access_key']
        bucket_name= self.connector_information['bucket_name']
        prefix = self.connector_information.get("prefix", "")
        session = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_access_key,
        )
        s3_resource = session.resource('s3')
        s3_client = session.client("s3")
        bucket = s3_resource.Bucket(bucket_name)

        # List out the files to be passed on
        for obj in bucket.objects.filter(Prefix=prefix):
            # Hard coded metadata available on the object
            metadata = {
                "last_modified" : obj.last_modified
            }
            selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
            # If metadata passed, then I will add all the user generated values that are associated to the file
            if "metadata" in self.selector.to_metadata:
                # Make an additional call to get the full context
                additional_metdata:dict = s3_client.head_object(Bucket=bucket_name, Key=obj.key)
                selected_metadata.update(additional_metdata['Metadata'])
            yield CloudFile(file_identifier=obj.key, metadata=selected_metadata, id=obj.key)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # Connect to S3
        aws_key_id= self.connector_information['aws_key_id']
        aws_access_key= self.connector_information['aws_access_key']
        bucket_name= self.connector_information['bucket_name']
        prefix = self.connector_information.get("prefix", "")   
        session = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_access_key,
        )
        s3_resource = session.resource('s3')
        s3_client = session.client("s3")
        bucket = s3_resource.Bucket(bucket_name)

        # List out the files that have changed
        for obj in bucket.objects.filter(Prefix=prefix):    
            # Check if file has changed
            if(last_run < obj.last_modified):
                # If file changed, then download
                metadata = {
                    "key" : obj.key,
                    "last_modified" : obj.last_modified
                }
                selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
                # If metadata passed, then I will add all the user generated values that are associated to the file
                if "metadata" in self.selector.to_metadata:
                    # Make an additional call to get the full context
                    additional_metdata:dict = s3_client.head_object(Bucket=bucket_name, Key=obj.key)
                    selected_metadata.update(additional_metdata['Metadata'])
                yield CloudFile(file_identifier=obj.key, metadata=selected_metadata, id=obj.key)

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        # Connect to S3
        aws_key_id= self.connector_information['aws_key_id']
        aws_access_key= self.connector_information['aws_access_key']
        bucket_name= self.connector_information['bucket_name']
        session = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_access_key,
        )
        s3_client = session.client("s3")
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = f"{temp_dir}/{cloudFile.file_identifier}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            s3_client.download_file(bucket_name, cloudFile.file_identifier, file_path)
            yield LocalFile(file_path=file_path, metadata=cloudFile.metadata, id=cloudFile.id)

    def validate(self) -> bool:
        try:
            aws_key_id= self.connector_information['aws_key_id']
            aws_access_key= self.connector_information['aws_access_key']
            bucket_name= self.connector_information['bucket_name']
        except:
            raise ValueError("Required properties not set")
        
        if not all(x in self.availableMetadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        try:
            session = boto3.Session(
                aws_access_key_id=aws_key_id,
                aws_secret_access_key=aws_access_key,
            )
            s3_resource = session.resource('s3')
            s3_client = session.client("s3")
            bucket = s3_resource.Bucket(bucket_name)
        except Exception as e:
            raise Exception(f"Connection to S3 failed, check key and key ID. See Exception: {e}")
        return True 




from datetime import datetime
from neumai.DataConnectors.DataConnector import DataConnector
from typing import List, Generator
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
import tempfile


class NeumFileConnector(DataConnector):
    """ Neum Simple File Connector \n
    connector_information required:[ url ] \n
    available metadata: [ url ]\n
    available content: [ file ]"""

    @property
    def connector_name(self) -> str:
        return "NeumSimpleFileConnector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["url"]

    @property
    def optionalProperties(self) -> List[str]:
        return []

    @property
    def availableMetadata(self) -> str:
        return ['url']

    @property
    def availableContent(self) -> str:
        return ['file']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        available_metadata = {'url':self.connector_information['url']}
        selected_metadata  = {k: available_metadata[k] for k in self.selector.to_metadata if k in available_metadata}
        yield CloudFile(file_identifier=self.connector_information['url'], metadata=selected_metadata, id=self.connector_information['url'])

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # Delta is not different, we are just getting one file. 
        yield self.connect_and_list_full()

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        # Connect to random file location
        import requests
        response = requests.get(cloudFile.file_identifier)

        # Download file
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_file.write(response.content)
            yield LocalFile(file_path=temp_file.name, metadata=cloudFile.metadata, id=cloudFile.id)

    def validate(self) -> bool:
        # Check for required properties
        try:
            url = self.connector_information['url']
        except:
            raise ValueError("Required properties not set")
        
        # Check for metadata
        if not all(x in self.availableMetadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided") 
        return True 



from datetime import datetime
from neumai.DataConnectors.DataConnector import DataConnector
from typing import List, Generator
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
import singlestoredb as s2
import json


class SingleStoreConnector(DataConnector):
    """SingleStore Connector \n
    connector_information requires:\n
    [ url, query ]"""
    
    @property
    def connector_name(self) -> str:
        return "SingleStoreConnector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["url", "query"]

    @property
    def optionalProperties(self) -> List[str]:
        return ["batch_size"]
    
    @property
    def availableMetadata(self) -> str:
        return []

    @property
    def availableContent(self) -> str:
        return ['row']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["NeumJSONLoader"]
    
    def datetime_serializer(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        raise TypeError("Type not serializable")

    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        url = self.connector_information['url']
        query = self.connector_information['query']
        batch_size = 1000
        if 'batch_size' in self.connector_information.keys():
            batch_size = self.connector_information['batch_size']
        
        with s2.connect(url, results_type="dict") as conn:
            with conn.cursor() as cur:
                batch_rows = []
                cur.execute(query)
                while True:
                    rows = cur.fetchmany(batch_size)
                    if not rows:
                        break

                    for row in rows:
                        batch_rows.append(row)
                        if(len(batch_rows) == batch_size):
                            json_object = json.dumps(batch_rows, default=self.datetime_serializer)
                            yield CloudFile(data=json_object, metadata={})
                            batch_rows = []

                if len(batch_rows) > 0:
                        json_object = json.dumps(batch_rows, default=self.datetime_serializer)
                        yield CloudFile(data=json_object, metadata={})

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # No metadatadata to determine what rows are new. Needs to be done through websocket
        yield from self.connect_and_list_full()

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
         data = json.loads(cloudFile.data)
         for row in data:
            yield LocalFile(in_mem_data=dict(row), metadata=cloudFile.metadata)
    
    def validate(self) -> bool:
        try:
            url = self.connector_information['url']
            query = self.connector_information['query']
        except:
            raise ValueError("Required properties not set")
        
        try: 
            s2.connect(url, results_type="dict")
        except:
            raise Exception("Connection URL is incorrect")
        return True 
                


from .TriggerSyncTypeEnum import TriggerSyncTypeEnum
from fastapi.responses import JSONResponse

class PipelineRunTaskDetails(object):
    def __init__(self, completed_embedding_tasks: int, completed_storing_tasks: int, failed_embedding_tasks: int, failed_storing_tasks: int ):
        self.completed_embedding_tasks = completed_embedding_tasks
        self.completed_storing_tasks = completed_storing_tasks
        self.failed_embedding_tasks = failed_embedding_tasks
        self.failed_storing_tasks = failed_storing_tasks

    def toJson(self):
        json_to_return = {}
        json_to_return['completed_embedding_tasks'] = self.completed_embedding_tasks
        json_to_return['completed_storing_tasks'] = self.completed_storing_tasks
        json_to_return['failed_embedding_tasks'] = self.failed_embedding_tasks
        json_to_return['failed_storing_tasks'] = self.failed_storing_tasks

        return json_to_return
    
    def as_pipeline_run_task_details(task_details: dict):
        if task_details == None:
            return None
        return PipelineRunStatus(
            completed_embedding_tasks=task_details.get("completed_embedding_tasks", None),
            completed_storing_tasks=task_details.get("completed_storing_tasks", None),
            failed_embedding_tasks=task_details.get("failed_embedding_tasks", None),
            failed_storing_tasks=task_details.get("failed_storing_tasks", None),
        )
    
class PipelineRunStatus(object):
    def __init__(self, status: str, message: str = None, exception_detail :str = None):
        self.status = status
        self.message = message
        self.exception_detail = exception_detail

    def toJson(self):
        json_to_return = {}
        json_to_return['status'] = self.status
        json_to_return['message'] = self.message
        json_to_return['exception_detail'] = self.exception_detail

        return json_to_return
    
    def as_pipeline_run_status(detailed_status: dict):
        if detailed_status == None:
            return PipelineRunStatus(status="Unknown",message="Something went wrong in our end, please email kevin@tryneum.com and we will get you unblocked")
        return PipelineRunStatus(
            status=detailed_status.get("status", None),
            message=detailed_status.get("message", None),
            exception_detail=detailed_status.get("exception_detail", None),
        )
    
class PipelineRun(object):
    def __init__(self, id: str, created: float, pipeline_id: str, trigger_type: str, sync_type: TriggerSyncTypeEnum, detailed_status:PipelineRunStatus = None, vectors_written=0, task_details: PipelineRunTaskDetails = None, last_updated:float = None):
        self.pipeline_id = pipeline_id
        self.id = id
        self.created = created
        self.trigger_type = trigger_type    
        self.detailed_status = detailed_status
        self.sync_type = sync_type
        self.vectors_written = vectors_written
        self.task_details = task_details
        self.last_updated = last_updated

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['id'] = self.id
        json_to_return['pipeline_id'] = self.pipeline_id
        json_to_return['created'] = self.created
        json_to_return['trigger_type'] = self.trigger_type
        json_to_return['sync_type'] = self.sync_type
        json_to_return['vectors_written'] = self.vectors_written
        json_to_return['last_updated'] = self.last_updated
        if self.task_details != None:
            json_to_return['task_details'] = self.task_details.toJson()
        if(self.detailed_status != None):
            json_to_return['detailed_status'] = self.detailed_status.toJson()
        else:
            json_to_return['detailed_status'] = {"status":"None", "message":"", "exception_detail":None}

        return json_to_return
                                                 
    def set_id(self, id: str):
        self.id = id

    def set_detailed_status(self, detailed_status: PipelineRunStatus):
        self.detailed_status = detailed_status
    
    def set_pipeline_id(self, pipeline_id: str):
        self.pipeline_id = pipeline_id

    def set_created(self, created: float):
        self.created = created
    
    def as_pipeline_run(dct:dict):
        if dct == None:
            return PipelineRun(id=None, created=None, pipeline_id = None, trigger_type=None, sync_type=None, detailed_status=None, vectors_written=None)
        return PipelineRun(
            id=dct.get("id", None),
            pipeline_id=dct.get("pipeline_id"),
            created=dct.get("created", None),
            trigger_type=dct.get("trigger_type", None),
            sync_type=TriggerSyncTypeEnum.as_trigger_sync_type(trigger_sync_type=dct.get("sync_type", None)),
            detailed_status=PipelineRunStatus.as_pipeline_run_status(detailed_status=dct.get("detailed_status", None)),
            vectors_written=dct.get("vectors_written", None),
            task_details=PipelineRunTaskDetails.as_pipeline_run_task_details(dct.get("task_details", None)),
            last_updated=dct.get("last_updated", None)
        )
    
class PipelineRunModel(JSONResponse):
    def __init__(
        self,
        content: PipelineRun,
        status_code: int = 200,
    ) -> None:
        super().__init__(content, status_code, headers=None, media_type="application/json", background=None)


from neumai.Loaders import (
    AutoLoader,
    MarkdownLoader,
    NeumJSONLoader,
    NeumCSVLoader,
    PDFLoader,
    HTMLLoader,
)
from neumai.Loaders.Loader import Loader
from neumai.Shared.Selector import Selector
from starlette.exceptions import HTTPException

def as_loader(dct:dict):
    if dct == None:
        raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
    loader_name = dct.get("loader_name", None)
    loader_information = dct.get("loader_information", None)
    selector = Selector.as_selector(dct.get("selector", None))
    if loader_name == "AutoLoader":
        return AutoLoader(loader_information=loader_information, selector=selector)
    elif loader_name == "MarkdownLoader":
        return MarkdownLoader(loader_information=loader_information, selector=selector)
    elif loader_name == "NeumJSONLoader":
        return NeumJSONLoader(loader_information=loader_information, selector=selector)
    elif loader_name == "NeumCSVLoader":
        return NeumCSVLoader(loader_information=loader_information, selector=selector)
    elif loader_name == "PDFLoader":
        return PDFLoader(loader_information=loader_information, selector=selector)
    elif loader_name == "HTMLLoader":
        return HTMLLoader(loader_information=loader_information, selector=selector)
    return None


from .Pipeline import Pipeline
from .Pipeline import Pipeline
from fastapi.responses import JSONResponse

class PipelineModel(JSONResponse):
    def __init__(
        self,
        content: Pipeline,
        status_code: int = 200,
    ) -> None:
        super().__init__(content, status_code, headers=None, media_type="application/json", background=None)

class PipelinesModel(JSONResponse):
    def __init__(
        self,
        content: dict,
        status_code: int = 200,
    ) -> None:
        super().__init__(content, status_code, headers=None, media_type="application/json", background=None)



from .CloudFile import CloudFile
from .LocalFile import LocalFile
from .NeumDocument import NeumDocument
from .NeumVector import NeumVector
from .Selector import Selector
from .NeumSinkInfo import NeumSinkInfo, NeumSinkInfoModel
from .NeumSearch import NeumSearchResult, NeumSearchResultModel


from neumai.Chunkers.Chunker import Chunker
from neumai.Shared.NeumDocument import NeumDocument
from typing import List, Generator
from langchain.text_splitter import (RecursiveCharacterTextSplitter)

class RecursiveChunker(Chunker):
    """Recursive Chunker \n
    chunker_information optional: \n
    [chunk_size, chunk_overlap, batch_size]"""
    
    @property
    def chunker_name(self) -> str:
        return "RecursiveChunker"
    
    @property
    def requiredProperties(self) -> List[str]:
        return []

    @property
    def optionalProperties(self) -> List[str]:
        return ["chunk_size" , "chunk_overlap" , "batch_size", "separators"]

    def chunk(self, documents:List[NeumDocument]) -> Generator[List[NeumDocument], None, None]:
        
        batch_size = self.chunker_information.get('batch_size', 1000)

        text_splitter = RecursiveCharacterTextSplitter(
            separators = self.chunker_information.get('separators', ["\n\n", "\n", " ", ""]),
            chunk_size = self.chunker_information.get('chunk_size', 500),
            chunk_overlap  = self.chunker_information.get('chunk_overlap', 0),
            length_function = len,
        )

        # Iterate through documents to chunk them and them merge them back up
        documents_to_embed:List[NeumDocument] = []
        for doc in documents:
            chunks = text_splitter.split_text(doc.content)
            for i in range(len(chunks)):
                documents_to_embed.append(NeumDocument(id=doc.id + "_" + str(i), content=chunks[i], metadata=doc.metadata))
                if(len(documents_to_embed) == batch_size):
                    yield documents_to_embed
                    documents_to_embed = []
                    
        # Pass last remaining
        if(len(documents_to_embed) > 0):
            yield documents_to_embed
    
    def validate(self) -> bool:
        return True   


from datetime import datetime
from neumai.DataConnectors.DataConnector import DataConnector
from typing import List, Generator
from supabase import create_client, Client
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
import tempfile
import os


class SupabaseConnector(DataConnector):
    """" Supabase File Connector \n
    connector_information requires:\n
    [ bucket, folder, url, key ]"""

    @property
    def connector_name(self) -> str:
        return "SupabaseConnector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["bucket","folder", "url", "key"]

    @property
    def optionalProperties(self) -> List[str]:
        return []
    
    @property
    def availableMetadata(self) -> str:
        return ['name', 'updated_at', 'created_at', 'last_accessed_at']

    @property
    def availableContent(self) -> str:
        return ['file']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        # Connect to supabase
        bucket = self.connector_information["bucket"]
        folder = self.connector_information['folder']
        url = self.connector_information['url']
        key = self.connector_information['key']
        supabase: Client = create_client(url, key)
        #Process files
        file_list = supabase.storage.from_(bucket).list(folder)
        for file in file_list:
            # Download each file
            name = file['name']
            selected_metadata  = {k: file[k] for k in self.selector.to_metadata if k in file}
            yield CloudFile(file_identifier=name, metadata=selected_metadata, id=name)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # Connect to supabase
        bucket = self.connector_information["bucket"]
        folder = self.connector_information['folder']
        url = self.connector_information['url']
        key = self.connector_information['key']
        supabase: Client = create_client(url, key)
        #Process files
        file_list = supabase.storage.from_(bucket).list(folder)
        for file in file_list:
            # Check if file has changed
            last_update_date = datetime.strptime(file['updated_at'], "%Y-%m-%dT%H:%M:%S.%fZ")
            if(last_run < last_update_date):
                #If file changed, then download
                name = file['name']
                selected_metadata  = {k: file[k] for k in self.selector.to_metadata if k in file}
                yield CloudFile(file_identifier=name, metadata=selected_metadata, id=name)

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        # Connect to supabase
        url = self.connector_information['url']
        key = self.connector_information['key']
        folder= self.connector_information['folder']
        bucket = self.connector_information["bucket"]

        supabase: Client = create_client(url, key)
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = f"{temp_dir}/{folder}/{cloudFile.file_identifier}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(f"{file_path}", "wb") as file:
                supabase_file = supabase.storage.from_(bucket).download(folder + "/" + cloudFile.file_identifier)
                file.write(supabase_file)
            yield LocalFile(file_path=file_path, metadata=cloudFile.metadata, id=cloudFile.id)

    def validate(self) -> bool:
        try:
            folder= self.connector_information['folder']
            url = self.connector_information['url']
            key = self.connector_information['key']
        except:
            raise ValueError("Required properties not set")
        
        if not all(x in self.availableMetadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        try:
            supabase: Client = create_client(url, key)
        except Exception as e:
            raise Exception(f"Connection to Supabase failed, check credentials. See Exception: {e}")
        return True 



from neumai.Chunkers import (
    CharacterChunker,
    CustomChunker,
    RecursiveChunker
)    
from starlette.exceptions import HTTPException

def as_chunker(dct:dict):
    if dct == None:
        raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
    chunker_name = dct.get("chunker_name", None)
    chunker_information = dct.get("chunker_information", None)
    if chunker_name == "CharacterChunker":
        return CharacterChunker(chunker_information=chunker_information)
    elif chunker_name == "CustomChunker":
        return CustomChunker(chunker_information=chunker_information)
    elif chunker_name == "RecursiveChunker":
        return RecursiveChunker(chunker_information=chunker_information)
    return None


from abc import abstractmethod, ABC
from typing import List, Generator
from neumai.Shared.NeumDocument import NeumDocument
from neumai.Shared.LocalFile import LocalFile
from neumai.Loaders.Loader import Loader
from neumai.Loaders.NeumCSVLoader import NeumCSVLoader
from neumai.Loaders.NeumJSONLoader import NeumJSONLoader
from neumai.Loaders.HTMLLoader import HTMLLoader
from neumai.Loaders.MarkdownLoader import MarkdownLoader
from neumai.Loaders.PDFLoader import PDFLoader
from langchain.document_loaders import UnstructuredFileLoader

class AutoLoader(Loader):
    """ Auto Loader\n loader_information contains: [] """

    @property
    def loader_name(self) -> str:
        return "AutoLoader"
    
    @property
    def requiredProperties(self) -> List[str]:
        return []

    @property
    def optionalProperties(self) -> List[str]:
        return []
    
    @property
    def availableMetadata(self) -> List[str]:
        return []

    @property
    def availableContent(self) -> List[str]:
        return []

    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        if "csv" in file.type:
            loader = NeumCSVLoader()
        elif "string" in file.type:
            yield NeumDocument(content=file.in_mem_data, metadata=file.metadata, id=file.id)
            return
        elif "application/octet-stream" in file.type:
            loader = MarkdownLoader()
        elif "html" in file.type:
            loader = HTMLLoader()
        elif "pdf" in file.type:
            loader = PDFLoader()
        elif "json" in file.type:
            loader = NeumJSONLoader()
        else:
            loader = UnstructuredFileLoader(file_path=file.file_path)
            documents = loader.load()
            for doc in documents:
                yield NeumDocument(id=file.id, content=doc.page_content, metadata=file.metadata)
            return 
        
        yield from loader.load(file=LocalFile)
    
    def validate(self) -> bool:
        return True   


from neumai.EmbedConnectors import (
    OpenAIEmbed,
    ReplicateEmbed,
    AzureOpenAIEmbed
)
from starlette.exceptions import HTTPException

def as_embed(dct:dict):
    if dct == None:
        raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
    embed_name = dct.get("embed_name", None)
    embed_information = dct.get("embed_information", None)
    if embed_name == "OpenAIEmbed":
        return OpenAIEmbed(embed_information=embed_information)
    elif embed_name == "ReplicateEmbed":
        return ReplicateEmbed(embed_information=embed_information)
    elif embed_name == "AzureOpenAIEmbed":
        return AzureOpenAIEmbed(embed_information=embed_information)
    return None